@@ -13,38 +13,42 @@ class ApoJWT():
         secret: Secret string used to encode and decode the access JWT
         exp_period: Length of time in seconds access tokens should be valid for. Default 900 (15 minutes) 
         iss: Issuer string used for additional security. Default ""
-        server_audience: Audience name of the server hosting the HTTP framework. 
+        server_audience: List of audiences named from the server hosting the HTTP framework. 
             Audience names are typically base address URLs. Ex: https://example.com
         algorithm: The algorithm to use when encoding/decoding. Default HS256
         admin_permission: Optional permission allowing full access to JWTs carrying this permission. USE CAREFULLY
         * token_finder: Function used to retrieve the access JWT from http request Authorization header. Default None
 
     Framework Configuration
-        asynch: Tells ApoJWT to use async decorators instead of the normal (FastAPI needs this True). Default False
+        async_framework: Tells ApoJWT to use async decorators instead of the normal (FastAPI needs this True). Default False
         * exception_handler: HTTP error handling function given an HTTP code and message as arguments
-
-    Utility
-        * permission_formatter: String formatting function that is given permission_name as an argument
-            Can be used to format request data in the permission name
     """
 
-    def __init__(self, secret: str, exp_period: int=900, iss: str="", asynch: bool=False, server_audience: str="default", algorithm: str="HS256", admin_permission: str=None, token_finder=None, permission_formatter=None, exception_handler=None):        
-        self.exp_period = int(exp_period)
-        self.refresh_exp_period = 0
-        self.iss = str(iss)
-        self.asynch = asynch
-        self.algorithm = algorithm
-
-        self.__token_f = token_finder
-        self.__permission_formatter = permission_formatter
-        self.__exception_handler = exception_handler
+    def __init__(self, secret: str, exp_period: int=900, iss: str="", server_audience: list=[], algorithm: str="HS256", admin_permission: str=None, template: str=None, async_framework: bool=False, token_finder=None, exception_handler=None):        
+        self.__exp_period = int(exp_period)
+        self.__refresh_exp_period = 0
+        self.__iss = str(iss)
+        self.__algorithm = algorithm
         self.__secret = str(secret)
-        self.__server_aud = server_audience
+        self.__server_aud = list(server_audience)
         self.__admin_perm = admin_permission
         self.__refresh_flag = False
         self.__refresh_f = None
         self.__refresh_secret = None
 
+        if template == "flask":
+            from flask_template import create_flask_template
+            self.__async = False
+            self.__token_f, self.__exception_handler = create_flask_template()
+        elif template == "fastapi":
+            from fastapi_template import create_fastapi_template 
+            self.__async = True
+            self.__token_f, self.__exception_handler = create_fastapi_template()
+        else:
+            self.__async = async_framework
+            self.__token_f = token_finder
+            self.__exception_handler = exception_handler
+
 
     def config_refresh(self, refresh_secret: str, refresh_exp_period: int=86400, refresh_finder=None):
         """Configures ApoJWT for use with refresh tokens
@@ -55,10 +59,9 @@ class ApoJWT():
         """
 
         self.__refresh_flag = True
-        self.__refresh_f = refresh_finder
         self.__refresh_secret = refresh_secret
-        self.refresh_exp_period = refresh_exp_period
-
+        self.__refresh_exp_period = refresh_exp_period
+        self.__refresh_f = refresh_finder
 
     def token_required(self, fn):
         """Verifies a JWT and all its claims
@@ -70,12 +73,21 @@ class ApoJWT():
             - invalid secret
             - invalid issuer
         """
-        if self.asynch is True:
+        if self.__async is True:
             @wraps(fn)
             async def wrapper(*args, **kwargs):
                 if self.__token_f is None:
                     raise TypeError("ApoJWT requires the token_finder attribute to be defined for validating JWTs")
-                token = self.__token_f(*args, **kwargs)
+                try:
+                    token = self.__token_f(*args, **kwargs)
+                except Exception as e:
+                    if self.__exception_handler is None:
+                        raise jwt.exceptions.InvalidTokenError(f"JWT is missing - Token Finder failed: {e}")
+                    else:
+                        kwarg_dict = dict(token_data=dict(), token_subject="", access_token="", code=401, msg=f"JWT is missing - Token Finder failed: {e}")
+                        self.__form_kwargs(self.__exception_handler, kwarg_dict, kwargs)
+                        self.__exception_handler(*args, **kwargs)
+                        raise NotImplementedError("ApoJWT Exception Handler must not return control")
                 token_dict = self.__decode(token, *args, **kwargs)
                 kwarg_dict = dict(token_data=token_dict["data"], token_subject=token_dict["sub"])
                 self.__form_kwargs(fn, kwarg_dict, kwargs)
@@ -86,7 +98,16 @@ class ApoJWT():
             def wrapper(*args, **kwargs):
                 if self.__token_f is None:
                     raise TypeError("ApoJWT requires the token_finder attribute to be defined for validating JWTs")
-                token = self.__token_f(*args, **kwargs)
+                try:
+                    token = self.__token_f(*args, **kwargs)
+                except Exception as e:
+                    if self.__exception_handler is None:
+                        raise jwt.exceptions.InvalidTokenError(f"JWT is missing - Token Finder failed: {e}")
+                    else:
+                        kwarg_dict = dict(token_data=dict(), token_subject="", access_token="", code=401, msg=f"JWT is missing - Token Finder failed: {e}")
+                        self.__form_kwargs(self.__exception_handler, kwarg_dict, kwargs)
+                        self.__exception_handler(*args, **kwargs)
+                        raise NotImplementedError("ApoJWT Exception Handler must not return control")
                 token_dict = self.__decode(token, *args, **kwargs)
                 kwarg_dict = dict(token_data=token_dict["data"], token_subject=token_dict["sub"])
                 self.__form_kwargs(fn, kwarg_dict, kwargs)
@@ -106,12 +127,21 @@ class ApoJWT():
             - invalid audience
         """
         def permission_decorated(fn):
-            if self.asynch is True:
+            if self.__async is True:
                 @wraps(fn)
                 async def wrapper(*args, **kwargs):
                     if self.__token_f is None:
                         raise TypeError("ApoJWT requires the token_finder attribute to be defined for validating JWTs")
-                    token = self.__token_f(*args, **kwargs)
+                    try:
+                        token = self.__token_f(*args, **kwargs)
+                    except Exception as e:
+                        if self.__exception_handler is None:
+                            raise jwt.exceptions.InvalidTokenError(f"JWT is missing - Token Finder failed: {e}")
+                        else:
+                            kwarg_dict = dict(token_data=dict(), token_subject="", access_token="", code=401, msg=f"JWT is missing - Token Finder failed: {e}")
+                            self.__form_kwargs(self.__exception_handler, kwarg_dict, kwargs)
+                            self.__exception_handler(*args, **kwargs)
+                            raise NotImplementedError("ApoJWT Exception Handler must not return control")
                     token_dict = self.__decode(token, *args, **kwargs)
                     decoded_permissions = token_dict["perms"]
                     if self.__admin_perm is not None and self.__admin_perm in decoded_permissions:
@@ -124,7 +154,7 @@ class ApoJWT():
                         if self.__exception_handler is None:
                             raise jwt.exceptions.InvalidTokenError(f"403: JWT is not authorized for this action")
                         else:
-                            kwarg_dict = dict(token_data=dict(), token_subject="", access_token="", code=403, exception_type="unauthorized", msg="JWT is not authorized for this action")
+                            kwarg_dict = dict(token_data=dict(), token_subject="", access_token="", code=403, msg="JWT is not authorized for this action")
                             self.__form_kwargs(self.__exception_handler, kwarg_dict, kwargs)
                             self.__exception_handler(*args, **kwargs)
                             raise NotImplementedError("ApoJWT Exception Handler must not return control")
@@ -137,7 +167,16 @@ class ApoJWT():
                 def wrapper(*args, **kwargs):
                     if self.__token_f is None:
                         raise TypeError("ApoJWT requires the token_finder attribute to be defined for validating JWTs")
-                    token = self.__token_f(*args, **kwargs)
+                    try:
+                        token = self.__token_f(*args, **kwargs)
+                    except Exception as e:
+                        if self.__exception_handler is None:
+                            raise jwt.exceptions.InvalidTokenError(f"JWT is missing - Token Finder failed: {e}")
+                        else:
+                            kwarg_dict = dict(token_data=dict(), token_subject="", access_token="", code=401, msg=f"JWT is missing - Token Finder failed: {e}")
+                            self.__form_kwargs(self.__exception_handler, kwarg_dict, kwargs)
+                            self.__exception_handler(*args, **kwargs)
+                            raise NotImplementedError("ApoJWT Exception Handler must not return control")
                     token_dict = self.__decode(token, *args, **kwargs)
                     decoded_permissions = token_dict["perms"]
                     if self.__admin_perm is not None and self.__admin_perm in decoded_permissions:
@@ -150,7 +189,9 @@ class ApoJWT():
                         if self.__exception_handler is None:
                             raise jwt.exceptions.InvalidTokenError(f"403: JWT is not authorized for this action")
                         else:
-                            self.__exception_handler(code=403, exception_type="unauthorized", msg="JWT is not authorized for this action", *args, **kwargs)
+                            kwarg_dict = dict(token_data=dict(), token_subject="", access_token="", code=403, msg="JWT is not authorized for this action", *args, **kwargs)
+                            self.__form_kwargs(self.__exception_handler, kwarg_dict, kwargs)
+                            self.__exception_handler(*args, **kwargs)
                             raise NotImplementedError("ApoJWT Exception Handler must not return control")
                     kwarg_dict = dict(token_data=token_dict["data"], token_subject=token_dict["sub"])
                     self.__form_kwargs(fn, kwarg_dict, kwargs)
@@ -160,57 +201,58 @@ class ApoJWT():
 
 
     def refresh(self, fn):
-        """Verifies a refresh token and returns a new access token"""
-        if self.asynch is True:
+        """Verifies a refresh token and returns the refresh data"""
+        if self.__async is True:
             @wraps(fn)
             async def wrapper(*args, **kwargs):
-                if self.__token_f is None:
-                    raise NotImplementedError("ApoJWT requires the token_finder attribute to be defined for refreshing JWTs")
                 if self.__refresh_f is None:
                     raise NotImplementedError("ApoJWT requires the refresh_finder attribute to be defined for refreshing JWTs")
-                if "token_subject" in kwargs.keys():
-                    kwargs.pop("token_subject")
-                if "token_data" in kwargs.keys():
-                    kwargs.pop("token_data")
-                ref_token = self.__refresh_f(*args, **kwargs)
-                self.__decode(ref_token, secret=self.__refresh_secret, *args, **kwargs)
-                acc_token = self.__token_f(*args, **kwargs)
-                token_payload = self.__decode(acc_token, verify_exp=False, *args, **kwargs)
-                token_payload["exp"] = int(time()) + self.exp_period
-                new_token = jwt.encode(token_payload, self.__secret, algorithm=self.algorithm)
-                kwarg_dict = dict(token_data=token_payload["data"], token_subject=token_payload["sub"], access_token=new_token)
+                if "refresh_data" in kwargs.keys():
+                    kwargs.pop("refresh_data")
+                try:
+                    ref_token = self.__refresh_f(*args, **kwargs)
+                except Exception as e:
+                    if self.__exception_handler is None:
+                        raise jwt.exceptions.InvalidTokenError(f"Refresh JWT is missing - Refresh Token Finder failed: {e}")
+                    else:
+                        kwarg_dict = dict(token_data=dict(), token_subject="", access_token="", code=401, msg=f"Refresh JWT is missing - Refresh Token Finder failed: {e}")
+                        self.__form_kwargs(self.__exception_handler, kwarg_dict, kwargs)
+                        self.__exception_handler(*args, **kwargs)
+                        raise NotImplementedError("ApoJWT Exception Handler must not return control")
+                refresh_payload = self.__decode(ref_token, secret=self.__refresh_secret, *args, **kwargs)
+                kwarg_dict = dict(refresh_data=refresh_payload["data"])
                 self.__form_kwargs(fn, kwarg_dict, kwargs)
                 return await fn(*args, **kwargs)
             return wrapper
         else:
             @wraps(fn)
             def wrapper(*args, **kwargs):
-                if self.__token_f is None:
-                    raise NotImplementedError("ApoJWT requires the token_finder attribute to be defined for refreshing JWTs")
                 if self.__refresh_f is None:
                     raise NotImplementedError("ApoJWT requires the refresh_finder attribute to be defined for refreshing JWTs")
-                if "token_subject" in kwargs.keys():
-                    kwargs.pop("token_subject")
-                if "token_data" in kwargs.keys():
-                    kwargs.pop("token_data")
-                ref_token = self.__refresh_f(*args, **kwargs)
-                self.__decode(ref_token, secret=self.__refresh_secret, *args, **kwargs)
-                acc_token = self.__token_f(*args, **kwargs)
-                token_payload = self.__decode(acc_token, verify_exp=False, *args, **kwargs)
-                token_payload["exp"] = int(time()) + self.exp_period
-                new_token = jwt.encode(token_payload, self.__secret, algorithm=self.algorithm)
-                kwarg_dict = dict(token_data=token_payload["data"], token_subject=token_payload["sub"], access_token=new_token)
+                if "refesh_data" in kwargs.keys():
+                    kwargs.pop("refresh_data")
+                try:
+                    ref_token = self.__refresh_f(*args, **kwargs)
+                except Exception as e:
+                    if self.__exception_handler is None:
+                        raise jwt.exceptions.InvalidTokenError(f"Refresh JWT is missing - Refresh Token Finder failed: {e}")
+                    else:
+                        kwarg_dict = dict(token_data=dict(), token_subject="", access_token="", code=401, msg=f"Refresh JWT is missing - Refresh Token Finder failed: {e}")
+                        self.__form_kwargs(self.__exception_handler, kwarg_dict, kwargs)
+                        self.__exception_handler(*args, **kwargs)
+                        raise NotImplementedError("ApoJWT Exception Handler must not return control")
+                refresh_payload = self.__decode(ref_token, secret=self.__refresh_secret, *args, **kwargs)
+                kwarg_dict = dict(refresh_data=refresh_payload["data"])
                 self.__form_kwargs(fn, kwarg_dict, kwargs)
                 return fn(*args, **kwargs)
             return wrapper
         
 
-    def create_token(self, sub: str="", permissions: list[str]=[], aud: list[str]=[], data: dict=dict(), refresh_data: dict=dict()):
+    def create_token(self, sub: str="", permissions: list[str]=[], data: dict=dict(), refresh_data: dict=dict()):
         """Encodes and returns an access JWT and optionally a refresh JWT
 
         sub: Subject of the JWT (typically some reference to the user of JWT)
         permissions: List of permissions to assign to token
-        aud: List of audiences token should be accepted by
         data: Any additional information that is needed
         refresh_data: If refresh is configured, this additional data is stored with the refresh token
 
@@ -221,49 +263,48 @@ class ApoJWT():
             - aud: Audience
             - iat: Issued At
         """
-        exp = int(time()) + self.exp_period
-        aud.append(self.__server_aud)
+        exp = int(time()) + self.__exp_period
         payload = {
             "exp": int(exp),
             "sub": str(sub),
             "nbf": datetime.now(tz=timezone.utc),
-            "iss": self.iss,
-            "aud": list(aud),
+            "iss": self.__iss,
+            "aud": list(self.__server_aud),
             "iat": datetime.now(tz=timezone.utc),
             "data": data,
             "perms": list(permissions)
         }
-        access_token = jwt.encode(payload, self.__secret, algorithm=self.algorithm)
+        access_token = jwt.encode(payload, self.__secret, algorithm=self.__algorithm)
         if self.__refresh_flag is True:
             if self.__refresh_secret is None:
                 raise ValueError("The refresh secret must be assigned using config_refresh")
-            ref_exp = int(time()) + self.refresh_exp_period
+            ref_exp = int(time()) + self.__refresh_exp_period
             refresh_payload = {
                 "exp": ref_exp,
                 "nbf": datetime.now(tz=timezone.utc),
-                "iss": self.iss,
-                "aud": list(aud),
+                "iss": self.__iss,
+                "aud": list(self.__server_aud),
                 "iat": datetime.now(tz=timezone.utc),
                 "data": refresh_data
             }
-            refresh_token = jwt.encode(refresh_payload, self.__refresh_secret, algorithm=self.algorithm)
+            refresh_token = jwt.encode(refresh_payload, self.__refresh_secret, algorithm=self.__algorithm)
             return access_token, refresh_token
         return access_token
 
 
-    def __decode(self, token, secret=None, verify_exp=True, *args, **kwargs):
+    def __decode(self, token, secret=None, *args, **kwargs):
         """Decodes the active jwt and returns the result"""
         not_implemented = NotImplementedError("ApoJWT Exception Handler must not return control")
         if secret is None:
             secret = self.__secret
         try:
-            decoded = jwt.decode(token, secret, issuer=self.iss, audience=self.__server_aud, algorithms=[self.algorithm], options={'verify_exp': verify_exp})
+            decoded = jwt.decode(token, secret, issuer=self.__iss, audience=self.__server_aud, algorithms=[self.__algorithm])
             return decoded
         except jwt.exceptions.InvalidSignatureError:
             if self.__exception_handler is None:
                 raise jwt.exceptions.InvalidSignatureError(f"401: JWT signature is invalid")
             else:
-                kwarg_dict = dict(token_data=dict(), token_subject="", access_token="", code=401, exception_type="signature", msg="JWT signature is invalid")
+                kwarg_dict = dict(token_data=dict(), token_subject="", access_token="", code=401, msg="JWT signature is invalid")
                 self.__form_kwargs(self.__exception_handler, kwarg_dict, kwargs)
                 self.__exception_handler(*args, **kwargs)
                 raise not_implemented
@@ -271,7 +312,7 @@ class ApoJWT():
             if self.__exception_handler is None:
                 raise jwt.exceptions.ExpiredSignatureError(f"403: JWT signature has expired")
             else:
-                kwarg_dict = dict(token_data=dict(), token_subject="", access_token="", code=403, exception_type="expired", msg="JWT signature has expired")
+                kwarg_dict = dict(token_data=dict(), token_subject="", access_token="", code=403, msg="JWT signature has expired")
                 self.__form_kwargs(self.__exception_handler, kwarg_dict, kwargs)
                 self.__exception_handler(*args, **kwargs)
                 raise not_implemented
@@ -279,7 +320,7 @@ class ApoJWT():
             if self.__exception_handler is None:
                 raise jwt.exceptions.InvalidIssuerError(f"401: JWT issuer is invalid")
             else:
-                kwarg_dict = dict(token_data=dict(), token_subject="", access_token="", code=401, exception_type="issuer", msg="JWT issuer is invalid")
+                kwarg_dict = dict(token_data=dict(), token_subject="", access_token="", code=401, msg="JWT issuer is invalid")
                 self.__form_kwargs(self.__exception_handler, kwarg_dict, kwargs)
                 self.__exception_handler(*args, **kwargs)
                 raise not_implemented
@@ -287,15 +328,15 @@ class ApoJWT():
             if self.__exception_handler is None:
                 raise jwt.exceptions.InvalidAudienceError(f"401: JWT audience is invalid")
             else:
-                kwarg_dict = dict(token_data=dict(), token_subject="", access_token="", code=401, exception_type="audience", msg="JWT audience is invalid")
+                kwarg_dict = dict(token_data=dict(), token_subject="", access_token="", code=401, msg="JWT audience is invalid")
                 self.__form_kwargs(self.__exception_handler, kwarg_dict, kwargs)
                 self.__exception_handler(*args, **kwargs)
                 raise not_implemented
-        except jwt.exceptions.InvalidTokenError:
+        except jwt.exceptions.InvalidTokenError as e:
             if self.__exception_handler is None:
                 raise jwt.exceptions.InvalidTokenError(f"401: JWT is invalid")
             else:
-                kwarg_dict = dict(token_data=dict(), token_subject="", access_token="", code=401, exception_type="invalid", msg="JWT is invalid")
+                kwarg_dict = dict(token_data=dict(), token_subject="", access_token="", code=401, msg=f"JWT is invalid: {e}")
                 self.__form_kwargs(self.__exception_handler, kwarg_dict, kwargs)
                 self.__exception_handler(*args, **kwargs)
                 raise not_implemented
